// Parallel Simulated Annealing
// for the Set-partitioning Problem
// Mateusz Styrczula, Michał Kraszewski
// 13.01.2009


#define _CRT_RAND_S

#include "Customer.h"
#include "Route.h"
#include "Solution.h"
#include "Main.h"
#include <algorithm>
#include <boost/optional.hpp>
#include <fstream>
#include <cstdlib>
#include <ctime>



namespace PSASP {



// ...
Solution::Solution (int map_size):
    base (0, 0),
    cost (0.0),
    map_size (map_size),
    route_random_access_table (map_size + 1, (boost::optional<std::list<Route>::iterator> ())) {}



// ...
Solution::Solution (const Solution &s):
    base (s.base),
    cost (s.cost),
    map_size (s.map_size),
    route_list (s.route_list),
    route_random_access_table (s.route_random_access_table.size ()) {
    int i = 0;
   
    for (std::list<Route>::iterator it = this->route_list.begin ();  it != this->route_list.end();  i++, it++)
        this->route_random_access_table[i] = boost::optional<std::list<Route>::iterator> (it);
}



// ...
void Solution::init () {
    Route curr_route (this->base);
    this->route_list.push_back (curr_route);
	std::list<Route>::iterator it = this->route_list.begin ();
	
	std::cout << "generating random clients." << std::endl;
	
	this->route_random_access_table[0] = boost::optional<std::list<Route>::iterator> (it);
    for (int i = 0;  i < this->map_size;  i++) {
    
        Customer c (0, 0);
        unsigned int srand_x, srand_y;
        rand_s (&srand_x); rand_s (&srand_y);
        c.x = (int)(srand_x % 200) - 100;
        c.y = (int)(srand_y % 200) - 100;
        
        if (!it->add (c)) {
            this->route_list.push_back (curr_route);
            it = this->route_list.end (); it--;
            it->add (c);
            this->route_random_access_table[i] = boost::optional<std::list<Route>::iterator> (it);
        }
    }
    this->compute_cost ();
    
    std::cout << "Initial cost: " << this->cost << std::endl;
    
	return;
}



// ...
void Solution::init (char *map_filename) {
    Route curr_route (this->base);
    this->route_list.push_back (curr_route);
	std::list<Route>::iterator it = this->route_list.begin ();
	
	std::ifstream input;
	std::vector<Customer> temp_cs;
	input.open (map_filename);
	bool input_available = true;
	if (!input) {
	    std::cout << "map.txt not available... generating random clients." << std::endl;
	    input_available = false;
	}
	int customer_amount = 0;
	input >> customer_amount;
	if (customer_amount != this->map_size  &&  input_available) {
	    std::cout << "map size (customer amount) mismatch... generating random clients." << std::endl;
	    input.close ();
	    input_available = false;
	}
	if (input_available) {
	    std::cout << "map.txt seems to be ok... using data from it." << std::endl;
    	for (int i = 0;  i < this->map_size;  i++) {
	        Customer c (0, 0);
	        input >> c.x >> c.y;
	        temp_cs.push_back (Customer (c.x, c.y));
	    }
        input.close ();
        std::random_shuffle (temp_cs.begin (), temp_cs.end ());
    }
	
	this->route_random_access_table[0] = boost::optional<std::list<Route>::iterator> (it);
    for (int i = 0;  i < this->map_size;  i++) {
    
        Customer c (0, 0);
        if (!input_available) {
            unsigned int srand_x, srand_y;
            rand_s (&srand_x); rand_s (&srand_y);
            c.x = (int)(srand_x % 200) - 100;
            c.y = (int)(srand_y % 200) - 100;
        } else {
            c.x = temp_cs[i].x;
            c.y = temp_cs[i].y;
        }
        
        if (!it->add (c)) {
            this->route_list.push_back (curr_route);
            it = this->route_list.end (); it--;
            it->add (c);
            this->route_random_access_table[i] = boost::optional<std::list<Route>::iterator> (it);
        }
    }
    this->compute_cost ();
    
    std::cout << "Initial cost: " << this->cost << std::endl;
    
	return;
}



// ...
const Solution &Solution::operator= (const Solution &s) {
    this->base = s.base;
    this->cost = s.cost;
    this->route_list = s.route_list;
    this->route_random_access_table.assign (s.route_random_access_table.size (), (boost::optional<std::list<Route>::iterator> ()));
    int i = 0;
    for (std::list<Route>::iterator it = this->route_list.begin ();  it != this->route_list.end();  i++, it++)
        this->route_random_access_table[i] = boost::optional<std::list<Route>::iterator> (it);
    return *this;
}



// ...
std::list<Route>::iterator Solution::get_random_route (bool potentially_with_new_empty_route, int *rat_id) {
    unsigned int tmp_rand = 0;
    rand_s (&tmp_rand);
    int r = tmp_rand % this->route_random_access_table.size ();
    boost::optional<std::list<Route>::iterator> p = this->route_random_access_table[r];
    
    if (potentially_with_new_empty_route  &&  r == 0) {
        while (p  &&  !((*p)->path.empty ())) {
            p = this->route_random_access_table[r = (r + 1) % this->route_random_access_table.size ()];
        }
        if (!p) {
            this->route_list.push_back (Route (this->base));
            std::list<Route>::iterator it = this->route_list.end (); it--;
            p = this->route_random_access_table[r] = boost::optional<std::list<Route>::iterator> (it);
        }
    } else {
        while (!p  ||  (*p)->path.empty ()) {
            p = this->route_random_access_table[r = (r + 1) % this->route_random_access_table.size ()];
        }
    }
    *rat_id = r;
    return *p;
}



// ...
double Solution::compute_cost () {
	this->cost = 0.0;
	for (std::list<Route>::const_iterator it = this->route_list.begin ();  it != this->route_list.end ();  it++) {
		this->cost += it->length ();
	}
	return this->cost;
}



// ...
double Solution::compute_cost (double delta) {
    this->cost += delta;
    
    return this->cost;
}



// ...
void Solution::print () const {
    std::cout << std::endl;
    std::cout << "-------   cost: " << this->cost << " --------- solution output: " << std::endl;
    for (std::list<Route>::const_iterator it = this->route_list.begin ();  it != this->route_list.end ();  it++) {
        std::cout << this->base.x << "    " << this->base.y << std::endl;
        for (std::list<Customer>::const_iterator cit = it->path.begin ();  cit != it->path.end ();  cit++) {
            std::cout << cit->x << "    " << cit->y << std::endl;
        }
        std::cout << this->base.x << "    " << this->base.y << std::endl;
        std::cout << std::endl;
    }
}



} // end namespace PSASP
