// Parallel Simulated Annealing
// for the Set-partitioning Problem
// Mateusz Styrczula, Michał Kraszewski
// 13.01.2009



#define _CRT_RAND_S

#include "Task.h"
#include "Route.h"
#include "Solution.h"
#include "Main.h"
#include "Customer.h"
#include <boost/thread/mutex.hpp>
#include <cstdlib>
#include <ctime>



namespace PSASP {



// ...
Task::Task (Main *m, int task_id):
    thread_running (false),
    main (m),
    task_id (task_id),
    iteration_counter (0),
    best_local_solution (m->map_size),
    current_solution (m->map_size) {
    this->best_local_solution = this->main->best_solution;
    this->current_solution = this->main->best_solution;
    this->start ();
}



// ...
Task::Task (const Task &t):
    thread_running (t.thread_running),
    task_id (t.task_id),
    thread (t.thread),
    main (t.main),
    current_solution (t.current_solution),
    best_local_solution (t.best_local_solution),
    iteration_counter (t.iteration_counter) {
}



// ...
const Task &Task::operator= (const Task &t) {
    this->thread_running = t.thread_running;
    this->task_id = t.task_id;
    this->thread = t.thread;
    this->main = t.main;
    this->current_solution = t.current_solution;
    this->best_local_solution = t.best_local_solution;
    this->iteration_counter = t.iteration_counter;
    return *this;
}



// ...
void Task::start () {
    this->thread = new boost::thread (*this);
    this->thread_running = true;
}



// ...
void Task::try_stop () {
    if (this->thread_running) {
        this->thread->join ();
        delete this->thread;
        this->thread_running = false;
    }
}



// ...
void Task::operator() () {
    this->body ();
}



// ...
void Task::body () {
    // introduce yourself
    this->main->cout_mutex.lock ();
    std::cout << "(" << this->task_id << ") Road Runner... beep beep!" << std::endl;
    this->main->cout_mutex.unlock ();
    
    // sync. before start
    this->main->main_barrier.wait ();
    
    // some variables
    int n_square = sqr (this->main->map_size);
    unsigned long missed, lucky_loose, hitted;
    missed = lucky_loose = hitted = 0;
    
    // disallow any input
    this->sending_mutex.lock ();
    
    // if you're not last task, ensure your input priority...
    if (this->task_id  !=  this->main->thread_count - 1)
        this->main->tasks[this->task_id + 1].receiving_mutex.lock ();

    // sync. before entering main loop
    this->main->help_barrier.wait ();
    
    do {
        for (this->iteration_counter = 0;  this->iteration_counter < n_square;  this->iteration_counter++) {
            this->annealing_step (&missed, &lucky_loose, &hitted);
            
            // propagate results in chain topology (if user demands it)
            if (this->main->prop_results    &&    this->iteration_counter % this->main->map_size  ==  0    &&    this->main->working) {
            
                // receiving
                if (this->task_id  !=  0) {
                    // allow input
                    this->sending_mutex.unlock ();
                    // passively wait for receiving operation completion
                    this->receiving_mutex.lock ();
                    // unlock receiving mutex
                    this->receiving_mutex.unlock ();
                    // disallow any input
                    this->sending_mutex.lock ();
                }
            
                // sending
                if (this->task_id  !=  this->main->thread_count - 1) {
                    // announce input to next thread (and passively wait for allowance)
                    this->main->tasks[this->task_id + 1].sending_mutex.lock ();
                    // yes we can -- process input
                    if (this->best_local_solution.cost  <  this->main->tasks[this->task_id + 1].best_local_solution.cost)
                        this->main->tasks[this->task_id + 1].best_local_solution = this->best_local_solution;
                    // input process finished
                    this->main->tasks[this->task_id + 1].sending_mutex.unlock ();
                    // unlock next thread
                    this->main->tasks[this->task_id + 1].receiving_mutex.unlock ();
                    // ensure your input priority (for next tete-a-tete)
                    this->main->tasks[this->task_id + 1].receiving_mutex.lock ();
                }

            }
            
        }
        
        // "visualize" progress
        std::cout << "." << std::flush;
        
        // lock main
        this->main->main_mutex.lock ();
        if (this->best_local_solution.cost  <  this->main->best_solution.cost) {
            this->main->best_solution = this->best_local_solution;
            this->main->equilibrium_counter = 0;
        }
        if (this->task_id == this->main->thread_count - 1) {
            this->main->temperature *= this->main->alpha;
            this->main->equilibrium_counter++;
            if (this->main->equilibrium_counter > Main::equilibrium_steps) {
                this->main->working = false;
            }
            // another "tiny" visualizer...
            std::cout << this->main->equilibrium_counter << std::flush;
        }
        // unlock main
        this->main->main_mutex.unlock ();
        
    } while (this->main->working);
    
    this->main->cout_mutex.lock ();
    if (this->task_id == this->main->thread_count - 1)
        std::cout << std::endl;
    std::cout << "(" << this->task_id << ") missed: " << missed << "   lucky_looses: " << lucky_loose << "   hitted: " << hitted << std::endl;
    this->main->cout_mutex.unlock ();
}



// ...
void Task::annealing_step (unsigned long *missed, unsigned long *lucky_loose, unsigned long *hitted) {
    double old_solution_cost = this->current_solution.cost;
    enum {option_01, option_02} breadcrumb;

    // select randomly a customer
    int route_01_rat_id;
    std::list<Route>::iterator route_01_pointer = this->current_solution.get_random_route (false, &route_01_rat_id);
    Customer &c = route_01_pointer->get_random_customer ();
    
    // select randomly a route (distinct from the customer's route selected above)
    std::list<Route>::iterator route_02_pointer;
    int route_02_rat_id;
    do {
        route_02_pointer = this->current_solution.get_random_route (true, &route_02_rat_id);
    } while (route_02_pointer == route_01_pointer);
    
    // ...
    double cost_delta;
    Customer *d = NULL;
    if (route_02_pointer->path.size ()  <  3) {
        route_02_pointer->add (c, &cost_delta);
        this->current_solution.compute_cost (cost_delta);
        route_01_pointer->remove (c, &cost_delta);
        this->current_solution.compute_cost (cost_delta);
        breadcrumb = option_01;
    } else {
        d = &(route_02_pointer->get_random_customer ());
        double routes_cost = route_01_pointer->length () + route_02_pointer->length ();
        c.exchange (*d);
        double new_routes_cost = route_01_pointer->recompute_length () + route_02_pointer->recompute_length ();
        cost_delta = new_routes_cost - routes_cost;
        this->current_solution.compute_cost (cost_delta);
        breadcrumb = option_02;
    }

    // ...
    double new_solution_cost = this->current_solution.cost;
    double delta = new_solution_cost - old_solution_cost;
    unsigned int tmp_rnd = 0;
    rand_s (&tmp_rnd);
    tmp_rnd = tmp_rnd % 1000000003;
    double x = (double)(tmp_rnd) / (1000000003.0);
    
    if (delta < 0.0  ||  x < (this->main->temperature) / (this->main->temperature + delta)) {
        if (new_solution_cost  <  this->best_local_solution.cost) {
            this->best_local_solution = this->current_solution;
            *hitted += 1;
        } else {
            *lucky_loose += 1;
        }
    } else {    // rollback changes...
        *missed += 1;
        if (breadcrumb == option_01) {
            route_01_pointer->add (c, &cost_delta);
            this->current_solution.compute_cost (cost_delta);
            route_02_pointer->remove (c, &cost_delta);
            this->current_solution.compute_cost (cost_delta);
        } else {
            d->exchange (c);
            route_01_pointer->recompute_length ();
            route_02_pointer->recompute_length ();
            this->current_solution.compute_cost ((-1) * cost_delta);
        }
    }

}    



} // end namespace PSASP
